#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <syslog.h>
#include "aclfsd.h"
#include <dlfcn.h>

static ACL_MODULE **tbmod;
static int nbmod;
static int maxmod;

static void *dummy_initctx (int, const char *[])
{
	return NULL;
}

static void dummy_endctx (void *)
{
}


PUBLIC ACL_MODULE::ACL_MODULE(const char *_name)
{
	name = strdup(_name);
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"/usr/lib/virtualfs/aclfsd/modules/%s.so",_name);
	handle = dlopen (path,RTLD_LAZY);
	if (handle == NULL){
		syslog (LOG_ERR,"Can't load module %s: %m",_name);
	}else{
		initctx = (initctx_f)dlsym (handle,"initctx");
		if (initctx == NULL) initctx = dummy_initctx;
		endctx = (endctx_f)dlsym (handle,"endctx");
		if (endctx == NULL) endctx = dummy_endctx;
		testctx = (testctx_f)dlsym (handle,"testctx");
	}
}

PUBLIC ACL_MODULE::~ACL_MODULE()
{
	free (name);
	if (handle != NULL) dlclose (handle);
}

PUBLIC bool ACL_MODULE::is_ok() const
{
	return handle != NULL;
}

PUBLIC void *ACL_MODULE::initcontext(int nbarg, const char *args[])
{
	return initctx (nbarg,args);
}

PUBLIC void ACL_MODULE::freecontext (void *stuff)
{
	endctx (stuff);
}

PUBLIC int ACL_MODULE::testcontext (
	void *stuff,	// Opaque data prepared by initcontext()
	ACL_CREDINFO &cinfo,	// Information about the client
	const char *dir,
	const char *file,
	OPER_CONTEXT &oper,
	int &rights)
{
	return testctx (stuff,cinfo,dir,file,oper,rights);
}
	

ACL_MODULE *aclmod_load (const char *name)
{
	ACL_MODULE *ret = NULL;
	if (strchr(name,'/')==NULL){
		for (int i=0; i<nbmod; i++){
			if (strcmp(tbmod[i]->name,name)==0){
				ret = tbmod[i];
				break;
			}
		}
		if (ret == NULL){
			ret = new ACL_MODULE (name);
			if (ret->is_ok()){
				if (nbmod == maxmod){
					maxmod += 100;
					tbmod = (ACL_MODULE**)realloc(tbmod,maxmod*sizeof(ACL_MODULE*));
					assert (tbmod != NULL);
				}
				tbmod[nbmod++] = ret;
			}else{
				delete ret;
				ret = NULL;
			}
		}
	}
	return ret;
}

PUBLIC ACL_MODULE_REF::ACL_MODULE_REF (const char *_name)
{
	name = strdup(_name);
	nbarg = 0;
	ctrl = NULL;
}


PUBLIC ACL_MODULE_REF::~ACL_MODULE_REF ()
{
	if (ctrl != NULL) ctrl->freecontext (stuff);
	for (int i=0; i<nbarg; i++) free (args[i]);
}

PUBLIC int ACL_MODULE_REF::addarg (const char *arg)
{
	int ret = -1;
	if (nbarg < MAX_MODARG){
		args[nbarg++] = strdup(arg);
		ret = 0;
	}
	return ret;
}

PUBLIC int ACL_MODULE_REF::check (
	ACL_CREDINFO &cinfo,	// Information about the client
	const char *dir,
	const char *file,
	OPER_CONTEXT &oper,
	int &rights)
{
	int ret = -1;
	if (ctrl == NULL){
		ctrl = aclmod_load (name);
		if (ctrl != NULL){
			stuff = ctrl->initcontext (nbarg,(const char**)args);
		}
	}
	if (ctrl != NULL){
		ret = ctrl->testcontext (stuff,cinfo,dir,file,oper,rights);
	}
	return ret;
}
