#include <string.h>
#include <stdio.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>
#include "aclfsd.h"

struct USER_SUPGRP{
	uid_t uid;
	gid_t tbgrp[MAX_SUPP_GROUP];
	int nbgrp;
	USER_SUPGRP *next;
	USER_SUPGRP (uid_t _uid, USER_SUPGRP *_next){
		uid = _uid;
		next = _next;
		nbgrp = 0;
	}
};

static USER_SUPGRP *tb[512];

/*
	Get the list of supplementary group for a user
*/
int acluser_getsupgrp (uid_t uid, gid_t tbgrp[MAX_SUPP_GROUP])
{
	int ret = -1;
	int hash = uid & 0x1ff;
	USER_SUPGRP *pt = tb[hash];
	while (pt != NULL){
		if (pt->uid == uid){
			ret = pt->nbgrp;
			if (ret > 0) memcpy (tbgrp,pt->tbgrp,ret*sizeof(gid_t));
			break;
		}
		pt = pt->next;
	}
	if (ret == -1){
		pt = new USER_SUPGRP (uid,tb[hash]);
		tb[hash] = pt;
		// Ok, we have to solve the information about this users
		struct passwd *p = getpwuid (uid);
		ret = 0;
		if (p != NULL){
			const int MAXSIZ_NAM=100;
			const int MAXSIZ_SCAN=100;
			char tbscan[MAXSIZ_SCAN][MAXSIZ_NAM+1];
			int noscan = 0;
			int nbscan = 1;
			strncpy(tbscan[0],p->pw_name,MAXSIZ_NAM);
			bool toomany = false;
			while (!toomany && noscan < nbscan){
				setgrent();
				struct group *g;
				const char *name = tbscan[noscan++];
				while (!toomany
					&& (g=getgrent())!=NULL){
					char **gr_mem = g->gr_mem;
					if (gr_mem != NULL){
						for (int i=0; gr_mem[i] != NULL; i++){
							if (strcmp(gr_mem[i],name)==0){
								if (ret == MAX_SUPP_GROUP-1){
									toomany = true;
									syslog(LOG_ERR,"User %s is member of too many group (max %d)"
										,p->pw_name,MAX_SUPP_GROUP);
								}else{
									tbgrp[ret++] = g->gr_gid;
									snprintf (tbscan[nbscan++],MAXSIZ_SCAN,"@%s"
										,g->gr_name);
									if (nbscan == MAXSIZ_SCAN){
										syslog(LOG_ERR,"Too many supplementary groups for user %s"
											,p->pw_name);
										toomany = true;
									}
								}
								break;
							}
						}
					}
				}
			}
			if (ret > 0){
				pt->nbgrp = ret;
				memcpy (pt->tbgrp,tbgrp,ret*sizeof(pt->tbgrp[0]));
			}
		}else{
			syslog(LOG_ERR,"No passwd entry for uid %u",uid);
		}
	}
	return ret;
}

/*
	Forget all information (supplementary groups) about users.
	It will be reloaded as needed.
*/
void acluser_reset ()
{
	for (int i=0; i<512; i++){
		USER_SUPGRP *pt = tb[i];
		while (pt != NULL){
			USER_SUPGRP *next = pt->next;
			delete pt;
			pt = next;
		}
		tb[i] = NULL;
	}
}

