/*
	Tell if a back-end is listening
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"

static void usage()
{
	fprintf (stderr,"litefs-ping version %s\n"
		"\n"
		"litefs-shutdown --sockets unix_socket [ unix_socket ]\n"
		,PACKAGE_REV);
}


static void fsig (int)
{
	// fprintf (stderr,"signal\n");
}

/*
	Install the signal handler and set the alarm
*/
static void cmdsock_setsig(int time_out)
{
	static struct sigaction siga;
	siga.sa_handler = fsig;
	sigaction (SIGALRM,&siga,NULL);
	alarm (time_out);
}

/*
	Remove the signal handler and reset the alarm
*/
static void cmdsock_resetsig ()
{
	static struct sigaction siga;
	siga.sa_handler = SIG_DFL;
	sigaction (SIGALRM,&siga,NULL);
	alarm (0);
}



static int litefs_ping (
	const char *sockpath,
	bool showerror,
	bool verbose)
{
	int ret = -1;
	cmdsock_setsig (5);
	int fd =  socket (AF_UNIX,SOCK_STREAM,0);
	if (fd == -1){
		if (showerror) perror("socket client");
	}else{
		struct sockaddr_un un;
		un.sun_family = AF_UNIX;
		strcpy (un.sun_path,sockpath);
		int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
		if (s == -1){
			if (showerror) perror("connect");
		}else{
			if (verbose) printf ("Pinging %s\n",sockpath);
			LITE_REQUEST req;
			req.cmd = LITE_VERSION;
			req.mode = LITEFS_PROTOCOL_VERSION;
			req.perm = 0;
			req.len  = 0;
			write (fd,&req,sizeof(LITE_REQUEST));
			LITE_REPLY rep;
			int len = read (fd,&rep,sizeof(rep));
			if (len == sizeof(LITE_REPLY)){
				if (rep.status == LITE_STATUS_OK){
					if (verbose) printf ("Ok\n");
					ret = 0;
				}else{
					fprintf (stderr,"%s listening, wrong version\n"
						,sockpath);
				}
			}
		}
		close (fd);
	}
	cmdsock_resetsig ();
	return ret;
}


int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		int verbose = 0;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--verbose")==0){
				verbose++;
			}else if (strcmp(argv[i],"--socket")==0){
				i++;
				ret = litefs_ping (argv[i],false,verbose);
			}else{
				usage();
			}
		}
	}
	return ret;
}	


