/*
	The fast pseudo-directory list all litefs-ssh mount point
	and allow a fast (no ssh/encryption) link to the same ressource.
	So you can do the following

	cd
	ls n/aserver
	ls n/fast/aserver

	and get the same result, except it is not crypted and is usually
	much faster.

	The two are linked. The fast link will use the slow (crypted/secure)
	link to establish is connection between litefs-proxy et litefsd
	using a non secure tcp connection. It uses tricks nevertheless
	to make this link fairly safe, although not crypted. It uses
	the following litefsd option, negociated over the secure link:

	--secret
	--oneconnect
	--findport
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <limits.h>
#include <dirent.h>
#include <ctype.h>
#include <pwd.h>
#include <stdlib.h>

static bool adminfs_checkext (const char *path, const char *ext)
{
	int lenext = strlen(ext);
	int lenpath = strlen(path);
	return lenpath >= lenext && strcmp(path+lenpath-lenext,ext)==0;
}
	
static bool adminfs_nosuffix(const char *path)
{
	bool ret = true;
	if (adminfs_checkext (path,".OLD")
		|| adminfs_checkext (path,".bak")
		|| adminfs_checkext (path,"~")){
		ret = false;
	}
	return ret;
}


int main (int argc, char *argv[])
{
	int ret = -1;
	const char *home = getenv ("HOME");
	if (home == NULL){
		fprintf (stderr,"No home directory, aborting\n");
	}else if ((argc == 2 || argc == 3) && strcmp(argv[1],"--list")==0){
		if (argc == 2){
			char path[PATH_MAX];
			// Prints the litefs-ssh entries in litefs.conf
			snprintf (path,sizeof(path)-1,"%s/.virtualfs/litefs.conf",home);
			FILE *fin = fopen (path,"r");
			if (fin != NULL){
				char buf[1000],mount[1000],tmp[1000],cmd[1000];
				while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
					if (sscanf (buf,"%s %s %s %s",mount,tmp,cmd,tmp)==4
						&& mount[0] != '#'
						&& strcmp(cmd,"litefs-ssh")==0){
						printf ("d %s %d %d 0\n",mount,getuid(),getgid());
					}
				}
				fclose (fin);
				ret = 0;
			}
			// Prints the adminfs entries as well. So fast contains both
			sprintf (path,"%s/.virtualfs/adminfs",home);
			DIR *dir = opendir (path);
			if (dir != NULL){
				struct dirent *ent;
				while ((ent=readdir(dir))!=NULL){
					if(adminfs_nosuffix(ent->d_name)){
						printf ("d %s %d %d 0\n",ent->d_name,getuid(),getgid());
					}
				}
				closedir (dir);
				ret = 0;
			}
		}else if (argc == 3){
			char path[PATH_MAX];
			sprintf (path,"%s/.virtualfs/fast/%s",home,argv[2]);
			struct stat st;
			if (stat(path,&st)==-1){
			}
		}
	}
	return ret;
}



